package examples;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Iterator;
import java.util.Set;

/**
 * Buffered SHOUTcast-Icecast audio stream
 */
final class ShoutcastIcecastStream extends InputStream {
	/** Connection timeout */
	private static final int CONNECTION_TIMEOUT_MS = 30000;
	/** Buffer size. The size must be enough to recognize a stream encoding */
	private static final int BUFFER_SIZE = 1 << 19;
	//
	private final HttpURLConnection mConnection;
	private final BufferedInputStream mInputStream;
	private final int mMetadataPeriod;
	private int mReadPosition = 0;
	private final byte[] mMetaData;
	//
	/**
	 * Creates stream to read SHOUTcast/Icecast stream with the specified buffer size and the metadata period
	 *
	 * @param in the underlying input stream.
	 * @param size the buffer size. Recommended size is 2^19.
	 * @param metadataPeriod
	 * @throws IllegalArgumentException - if size <= 0 or metadataPeriod < 0.
	 */
	public ShoutcastIcecastStream(final String url, final boolean isReadMetadata, final int size) throws IOException {
		if( ! url.toLowerCase().startsWith("http://") ) {
			throw new IllegalArgumentException("Error: not http url");
		}
		mConnection = (HttpURLConnection)new URL( url ).openConnection();
		mConnection.setRequestMethod("GET");
		//mConnection.setRequestProperty("User-Agent", APP_NAME );
		mConnection.setRequestProperty("Accept", "*/*");
		mConnection.setRequestProperty("Icy-MetaData", isReadMetadata ? "1" : "0");
		mConnection.setRequestProperty("Connection", "keep-alive");
		mConnection.setConnectTimeout( CONNECTION_TIMEOUT_MS );
		mConnection.setReadTimeout( CONNECTION_TIMEOUT_MS );
		mConnection.setUseCaches( false );
		mConnection.connect();
		final int ret = mConnection.getResponseCode();
		if( ret != HttpURLConnection.HTTP_OK ) {
			throw new IOException("Error: response code is " + ret );
		}
		// print response
		System.out.println("---- Server response:");
		final Set<String> keys = mConnection.getHeaderFields().keySet();
		for( final Iterator<String> it = keys.iterator(); it.hasNext(); ) {
			final String key = it.next();
			if( key != null ) {
				System.out.println( key + " " + mConnection.getHeaderField( key ) );
			}
		}
		System.out.println("----");
		//
		if( ! mConnection.getHeaderField("Content-Type").startsWith("audio/") &&
			! mConnection.getHeaderField("Content-Type").startsWith("application/ogg") ) {
			throw new IOException("Error: not a SHOUTcast / Icecast stream");
		}// icecast/shoutcast
		// System.out.println("SHOUTcast / Icecast protocol is detected");
		final int metadata_period = mConnection.getHeaderFieldInt( "icy-metaint", 0 );
		if( metadata_period < 0 ) {
			throw new IOException("Error: metadata period = " + metadata_period );
		}
		mMetadataPeriod = metadata_period;
		byte[] b = null;
		if( metadata_period != 0 ) {
			b = new byte[255 * 16];
		}
		mMetaData = b;
		mInputStream = new BufferedInputStream( mConnection.getInputStream(), size >= BUFFER_SIZE ? size : BUFFER_SIZE );
	}

	@Override
	public void close() throws IOException {
		mConnection.disconnect();
		mInputStream.close();
	}

	@Override
	public final boolean markSupported() {
		return true;
	}
	@Override
	public synchronized final void mark(final int readlimit) {
		mInputStream.mark( readlimit );
	}
	@Override
	public synchronized final void reset() throws IOException {
		mInputStream.reset();
		mReadPosition = 0;
	}

	@Override
	public final int available() throws IOException {
		return mInputStream.available();
	}

	@Override
	public synchronized final int read() throws IOException {
		if( mMetadataPeriod == 0 ) {
			return mInputStream.read();
		}
		if( mReadPosition >= mMetadataPeriod ) {
			// Get metadata length, max size is 255 * 16 = 4080 bytes without zero
			int length = mInputStream.read();
			if( length < 0 ) {
				return length;
			}
			length <<= 4;// size of the metadata
			if( length > 0 ) {
				int off = 0;
				int read;
				while( (read = mInputStream.read( mMetaData, off, length - off )) > 0 ) {
					off += read;
				}
				try {
					System.out.println( new String( mMetaData, 0, length, "UTF-8") );
				} catch(final Exception e) {
					System.out.println( new String( mMetaData, 0, length ) );
				}
			}
			mReadPosition = 0;
		}
		mReadPosition++;
		return mInputStream.read();
	}
	@Override
	public synchronized final int read(final byte[] b, int off, int len) throws IOException {
		if( mMetadataPeriod == 0 ) {
			return mInputStream.read( b, off, len );
		}
		int data = mMetadataPeriod - mReadPosition;
		if( data < len ) {
			int read = mInputStream.read( b, off, data );
			if( read < data ) {
				mReadPosition += read > 0 ? read : 0;
				return read;
			}
			off += data;
			len -= data;
			// Get metadata length, max size is 255 * 16 = 4080 bytes without zero
			int length = mInputStream.read();
			if( length < 0 ) {
				return length;
			}
			length <<= 4;// size of the metadata
			if( length > 0 ) {
				int offset = 0;
				int r;
				while( (r = mInputStream.read( mMetaData, offset, length - offset )) > 0 ) {
					offset += r;
				}
				try {
					System.out.println( new String( mMetaData, 0, length, "UTF-8") );
				} catch(final Exception e) {
					System.out.println( new String( mMetaData, 0, length ) );
				}
			}
			data = mMetadataPeriod;
			while( data < len ) {
				int r = mInputStream.read( b, off, data );
				if( r < data ) {
					if( r < 0 ) {
						r = 0;// buffer has some data. eof will be returns while next read
					}
					read += r;
					mReadPosition = r;
					return read;
				}
				read += r;
				off += data;
				len -= data;
				// Get metadata length, max size is 255 * 16 = 4080 bytes without zero
				length = mInputStream.read();
				if( length < 0 ) {
					return length;
				}
				length <<= 4;// size of the metadata
				if( length > 0 ) {
					int offset = 0;
					while( (r = mInputStream.read( mMetaData, offset, length - offset )) > 0 ) {
						offset += r;
					}
					try {
						System.out.println( new String( mMetaData, 0, length, "UTF-8") );
					} catch(final Exception e) {
						System.out.println( new String( mMetaData, 0, length ) );
					}
				}
			}
			int r = mInputStream.read( b, off, len );
			if( r < 0 ) {
				r = 0;// buffer has some data. eof will be returns while next read
			}
			read += r;
			mReadPosition = r;
			return read;
		}
		final int read = mInputStream.read( b, off, len );
		mReadPosition += read > 0 ? read : 0;
		return read;
	}
}
